/**
 * The MIT License (MIT)
 *
 * Copyright (c) 2021 Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 * Copyright (c) 2021 Abrar (https://gitlab.com/s96Abrar)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <QDebug>
#include <QObject>
#include <QWindow>
#include <QCoreApplication>
#include <wayland-client.h>

#include <private/qwaylandwindow_p.h>

#include "wayqt/WayQtUtils.hpp"
#include "wayqt/SessionLock.hpp"
#include "SessionLockImpl.hpp"

#include "ext-session-lock-v1-client-protocol.h"

WQt::SessionLockManager::SessionLockManager( ext_session_lock_manager_v1 *lockMgr ) {
    mObj = lockMgr;
}


WQt::SessionLockManager::~SessionLockManager() {
    ext_session_lock_manager_v1_destroy( mObj );
}


WQt::SessionLocker *WQt::SessionLockManager::getLocker() {
    qCritical() << "Creating locker";
    ext_session_lock_v1 *lock = ext_session_lock_manager_v1_lock( mObj );

    return new WQt::SessionLocker( lock );
}


ext_session_lock_manager_v1 *WQt::SessionLockManager::get() {
    return mObj;
}


WQt::SessionLocker::SessionLocker( ext_session_lock_v1 *sessLock ) {
    mObj = sessLock;
    qCritical() << "Session Lock: " << mObj;

    /** Start listening to the events of ext_session_lock_v1 */
    if ( wl_proxy_get_listener( (wl_proxy *)mObj ) != &mListener ) {
        ext_session_lock_v1_add_listener( mObj, &mListener, this );
    }
}


WQt::SessionLocker::~SessionLocker() {
    ext_session_lock_v1_destroy( mObj );
}


void WQt::SessionLocker::setup() {
    if ( mIsSetup == false ) {
        mIsSetup = true;

        if ( mPendingEvents.acquired ) {
            emit lockAcquired();
        }

        else if ( mPendingEvents.failed ) {
            emit lockFailed();
        }
    }
}


bool WQt::SessionLocker::isLocked() {
    return mLocked;
}


WQt::SessionLockSurface * WQt::SessionLocker::lockSurfaceForWindow( QWindow *window ) {
    QtWaylandClient::QWaylandWindow *waylandWindow = dynamic_cast<QtWaylandClient::QWaylandWindow *>( window->handle() );

    if ( waylandWindow == nullptr ) {
        qCritical() << "Unable to construct QtWaylandClient::QWaylandWindow* instance from QWindow*";
        return nullptr;
    }

    SessionLockSurfaceImpl *surface = qobject_cast<SessionLockSurfaceImpl *>( waylandWindow->shellSurface() );

    if ( surface == nullptr ) {
        qCritical() << "Unable to construct SessionLockSurfaceImpl* instance from QtWaylandClient::QWaylandWindow::shellSurface()";
        return nullptr;
    }

    return new SessionLockSurface( surface );
}


WQt::SessionLockSurface * WQt::SessionLocker::getLockSurface( QWindow *window, wl_output *output ) {
    QtWaylandClient::QWaylandWindow *waylandWindow = dynamic_cast<QtWaylandClient::QWaylandWindow *>( window->handle() );

    if ( waylandWindow == nullptr ) {
        qCritical() << "Unable to construct QtWaylandClient::QWaylandWindow* instance from QWindow*";
        return nullptr;
    }

    wl_surface *wlSurf = WQt::Utils::wlSurfaceFromQWindow( window );

    /* We need a valid wl_surface */
    if ( wlSurf == nullptr ) {
        return nullptr;
    }

    ext_session_lock_surface_v1 *lockSurf = ext_session_lock_v1_get_lock_surface( mObj, wlSurf, output );

    if ( lockSurf == nullptr ) {
        qCritical() << "Unable to construct SessionLockSurfaceImpl* instance from QtWaylandClient::QWaylandWindow::shellSurface()";
        return nullptr;
    }

    return new SessionLockSurface( window, lockSurf );
}


void WQt::SessionLocker::unlockAndDestroy() {
    if ( mLocked ) {
        ext_session_lock_v1_unlock_and_destroy( mObj );
    }
}


ext_session_lock_v1 *WQt::SessionLocker::get() {
    return mObj;
}


void WQt::SessionLocker::handleFinished( void *data, ext_session_lock_v1 * ) {
    WQt::SessionLocker *lock = reinterpret_cast<WQt::SessionLocker *>( data );

    /** Inform the listeners that the acquired lock is now destroyed */
    if ( lock->mIsSetup ) {
        if ( lock->mLocked ) {
            emit lock->lockFinished();
        }

        else {
            emit lock->lockFinished();
        }
    }

    else {
        lock->mPendingEvents.acquired = false;
        lock->mPendingEvents.failed   = true;
    }

    ext_session_lock_v1_destroy( lock->mObj );
    lock->mObj    = nullptr;
    lock->mLocked = false;
}


void WQt::SessionLocker::handleLocked( void *data, ext_session_lock_v1 * ) {
    WQt::SessionLocker *lock = reinterpret_cast<WQt::SessionLocker *>( data );

    lock->mLocked = true;

    if ( lock->mIsSetup ) {
        /** Inform the listeners that we have obtained the lock */
        emit lock->lockAcquired();
    }

    else {
        /** Store it in pending */
        lock->mPendingEvents.acquired = true;
    }
}


const ext_session_lock_v1_listener WQt::SessionLocker::mListener = {
    handleLocked,
    handleFinished,
};


WQt::SessionLockSurface::SessionLockSurface( QWindow *win, ext_session_lock_surface_v1 *lockSurf ) {
    QtWaylandClient::QWaylandWindow *waylandWindow = dynamic_cast<QtWaylandClient::QWaylandWindow *>( win->handle() );

    if ( waylandWindow != nullptr ) {
        /** make it visible, if it's not already visible */
        waylandWindow->setVisible( true );

        impl = new SessionLockSurfaceImpl( waylandWindow, lockSurf );
    }
}


WQt::SessionLockSurface::SessionLockSurface( SessionLockSurfaceImpl *impl ) {
    this->impl = impl;
}


WQt::SessionLockSurface::~SessionLockSurface() {
    delete impl;
    impl = nullptr;
}


bool WQt::SessionLockSurface::isValid() {
    return ( impl != nullptr );
}


ext_session_lock_surface_v1 *WQt::SessionLockSurface::get() {
    if ( impl == nullptr ) {
        return nullptr;
    }

    return impl->get();
}
